/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.util.Comparator;
import java.util.Spliterator;
import java.util.stream.Stream;

/**
 * <p>The root interface of the Java Array Framework. Arrays provide a layer on top of language level arrays. The main
 * goal of the framework is to provide an alternative to the Java Collection Framework for primitive arrays.</p>
 *
 * <p>The framework consists of three types of arrays, <i>immutable</i>, fixed-size <i>mutable</i>
 * and <i>dynamic</i> arrays. Immutable arrays protect their data by creating defensive copies on creation and
 * retrieval of elements. Mutable arrays expose the backing array and provides {@code set} method to modify elements
 * within this array. The backing array reference is constant for mutable arrays, hence it's size is constant as well.
 * An array is mutable if it implements {@link MutableArray}. In this case it must also return {@code true} for
 * {@code hasCharacteristics(MUTABLE)}.
 * Dynamic arrays allow arrays to grow and shrink by exposing {@code add},
 * {@code remove} and {@code clear} methods. The backing array reference of dynamic arrays is not constant over time,
 * the reference is replaced when the array grows or shrinks. An array is dynamic it implements {@link DynamicArray}.
 * In this case it must also return {@code true} for {@code hasCharacteristics(DYNAMIC)}.</p>
 *
 * <p>Arrays are constructed by the static interface methods that are declared by the typed interfaces of the array
 * hierarchy. Each of these static methods provide different semantics, such as {@code newInstance} for constructing
 * empty dynamic-size arrays, {@code valueOf} for wrapping {@code Array} around native arrays and {@code copyOf} for
 * copying the contents of native arrays or framework {@code Array}s.
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * Characteristics are always invalidated if conditions of these characteristics are broken as result of array
 * modification. Characteristics are only added in scenarios in which their conditions can be 'tested' without
 * traversal of elements. For example, the {@code SORTED} characteristic is added if the length of an array is reduced
 * to zero or one element, or if all elements are replaced by an array that is marked as being sorted.
 * The semantics of characteristics processing is documented for each "destructive" method. All implementations returned
 * by the framework adhere to these semantics. Custom implementations may choose to set characteristics at their
 * discretion as long as they do not violate the contract of these characteristics.
 * Users can take full control of characteristics assignment by means of the {@link MutableArray#doAction(ArrayAction)}
 * method.</p>
 *
 * <p>The Java Array Framework is tightly linked to the Java Collection Framework. While it might be desirable to have
 * {@code Array} extend the {@code Collection} interface, it does not because the {@code toArray} methods of
 * {@code Collection} would disallow the array to returning native primitive arrays. Furthermore, the array hierarchy
 * promotes defensive programming by enforcing (im)mutability at compile-time, in contrast to the optional
 * "destructive" methods of {@code Collection}, which merely act at run-time. Full interoperability of the array and
 * collection domain is achieved by the following bridging methods: {@link Arrays#copyOf(java.util.Collection)},
 * {@link Arrays#newArray(java.util.List)}, {@link Arrays#newMutableArray(java.util.List)},
 * {@link Arrays#newDynamicArray(java.util.List)}, {@link Arrays#asList(Array)},
 * {@link Arrays#asNavigableSet(DynamicArray)} and
 * {@link Arrays#asNavigableSet(DynamicObjectArray, java.util.Comparator)}.</p>
 *
 * @param <E> element type.
 * @author Leon van Zantvoort
 */
public interface Array<E> extends Iterable<E>, ArrayCharacteristics {

    /**
     * Returns the component type of the backing array.
     *
     * @return component type of the backing array.
     */
    Class<? super E> getComponentType();

    /**
     * Gets the comparator that is used for sorting this array, {@code null} if array is primitive array, or if array
     * sorted according to natural ordering of elements, or {@code null} if array is not sorted.
     *
     * @return a {@code Comparator}, or {@code null} if the elements are sorted in the natural order, or {@code null} if
     * array is not sorted.
     */
    @CheckForNull Comparator<? super E> getComparator();

    /**
     * <p>Returns the number of elements of this array. All index related methods must respect the boundary represented
     * by this value.</p>
     *
     * <p>Note that for mutable arrays, the backing array - as returned by {@code MutableArray.array()} -
     * may be larger.</p>
     *
     * @return length of array.
     */
    int length();

    /**
     * <p>Returns the offset relative to the backing array.</p>
     *
     * @return offset relative to the backing array.
     */
    int offset();

    /**
     * Returns {@code true} if this array contains no elements.
     *
     * @return {@code true} if this array contains no elements.
     */
    boolean isEmpty();

    /**
     * Returns the object found at {@code [offset + index]} of the backing array.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    E get(int index);

    /**
     * Returns {@code true} if this array contains the specified object. This method supports boxed primitives. This
     * means that if this array represents a primitive array, this method returns {@code true} if this array contains
     * the unboxed version of specified object.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object object whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified object.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if this array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with object.
     */
    boolean contains(Object object);

    /**
     * Returns {@code true} if this array contains all of the elements in the specified {@code array}. Primitives and
     * their object counterparts are regarded equal in this respect.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array is sorted, otherwise all elements are
     * traversed until the objects in specified array are found.</p>
     *
     * @param array array to be checked for containment in this array.
     * @return {@code true} if this array contains all of the elements in the specified array.
     * @throws NullPointerException if the specified array is null, or if the specified array contains null elements and
     * array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with elements.
     */
    boolean containsAll(@Nonnull Array<?> array);

    /**
     * Returns the index of the first occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with object.
     */
    int indexOf(Object object);

    /**
     * Returns the index of the last occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified object is found.</p>
     *
     * @param object object whose highest index in this array is to be returned.
     * @return highest index of object in array or {@code -1} if value is not found.
     * @throws NullPointerException if object is null and array is sorted by a comparator that does not permit null arguments.
     * @throws ClassCastException if this array is sorted and array contains non {@code Comparable} elements or comparator
     * not compatible with object.
     */
    int lastIndexOf(Object object);

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    ArrayIterator<E> iterator();

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    ArrayIterator<E> iterator(int index);

    /**
     * Returns an native array of the elements specified by {@code this} array. This method always returns a defensive
     * (range) copy of the backing array, even if this array is mutable.
     *
     * @return range copy of backing array.
     */
    Object toArray();

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    Array<E> offset(int offset);

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    Array<E> length(int length);

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    Array<E> subArray(int fromIndex, int toIndex);

    /**
     * Creates a {@code Spliterator} over the elements in this array.
     *
     * <p>The {@code Spliterator} reports:
     * <ul>
     *     <li>{@code Spliterator.SIZED};</li>
     *     <li>{@code Spliterator.SUBSIZED};</li>
     *     <li>{@code Spliterator.ORDERED};</li>
     *     <li>{@code Spliterator.IMMUTABLE} if {@code hasCharacteristics(MUTABLE)} returns {@code false};</li>
     *     <li>{@code Spliterator.NONNULL} if {@code hasCharacteristics(NONNULL)} returns {@code true};</li>
     *     <li>{@code Spliterator.SORTED} if {@code hasCharacteristics(SORTED)} returns {@code true};</li>
     *     <li>{@code Spliterator.DISTINCT} if {@code hasCharacteristics(DISTINCT)} returns {@code true}.</li>
     * </ul>
     *
     * <p>The default implementation is equivalent to, for this {@code array}:</p>
     * <pre>
     * return Arrays.spliterator(array);
     * </pre>
     *
     * @return a {@code Spliterator} over the elements in this array.
     */
    default Spliterator<E> spliterator() {
        return Arrays.spliterator(this);
    }

    /**
     * Returns a sequential {@code Stream} with this array as its source.
     *
     * <p>The default implementation is equivalent to, for this {@code array}:</p>
     * <pre>
     * return Arrays.stream(array);
     * </pre>
     *
     * @return a sequential {@code Stream} over the elements in this array.
     */
    default Stream<E> stream() {
        return Arrays.stream(this);
    }

    /**
     * Returns a parallel {@code Stream} with this array as its source.
     *
     * <p>The default implementation is equivalent to, for this {@code array}:</p>
     * <pre>
     * return Arrays.parallelStream(array);
     * </pre>
     *
     * @return a parallel {@code Stream} over the elements in this array.
     */
    default Stream<E> parallelStream() {
        return Arrays.parallelStream(this);
    }
}
