/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.AbstractList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.NavigableSet;
import java.util.Objects;
import java.util.RandomAccess;
import java.util.Set;
import java.util.SortedSet;
import java.util.Spliterator;
import java.util.function.Consumer;
import java.util.function.ToIntFunction;
import java.util.function.UnaryOperator;
import java.util.stream.Stream;

import static java.lang.reflect.Array.get;
import static java.lang.reflect.Array.newInstance;
import static org.squarebrackets.ArrayCharacteristics.DISTINCT;
import static org.squarebrackets.ArrayCharacteristics.SORTED;

/**
 * Bridges between Array and Collection domain.
 *
 * @author Leon van Zantvoort
 */
final class ArrayBridge {

    /**
     * <p>Returns an immutable array backed by the specified {@code list}. All operations on the resulting array are
     * delegated to the list.</p>
     *
     * <p>The list does not need to be empty at the time this method is invoked. The list must not be modified directly
     * after this method returns.</p>
     *
     * @param <E> element type.
     * @param list the list by which the array will be backed.
     * @return an immutable array backed by the specified list.
     * @throws NullPointerException if the specified list is null.
     */
    static <E> ObjectArray<E> newArray(@Nonnull List<E> list) {
        Objects.requireNonNull(list);
        return new ListArrayImpl<>(list, 0, list.size(), 0, null);
    }

    /**
     * <p>Returns a mutable array backed by the specified {@code list}. All operations on the resulting array are
     * delegated to the list.</p>
     *
     * <p>The list does not need to be empty at the time this method is invoked. The list must not be modified directly
     * after this method returns.</p>
     *
     * @param <E> element type.
     * @param list the list by which the array will be backed.
     * @return a mutable array backed by the specified list.
     * @throws NullPointerException if the specified list is null.
     */
    static <E> MutableObjectArray<E> newMutableArray(@Nonnull List<E> list) {
        Objects.requireNonNull(list);
        return new MutableListArrayImpl<>(list, 0, list.size(), 0, null);
    }

    /**
     * <p>Returns a dynamic array backed by the specified {@code list}. All operations on the resulting array are
     * delegated to the list.</p>
     *
     * <p>The list does not need to be empty at the time this method is invoked. The list must not be modified directly
     * after this method returns.</p>
     *
     * @param <E> element type.
     * @param list the list by which the array will be backed.
     * @return a dynamic array backed by the specified list.
     * @throws NullPointerException if the specified list is null.
     */
    static <E> DynamicObjectArray<E> newDynamicArray(@Nonnull List<E> list) {
        Objects.requireNonNull(list);
        return new DynamicListArrayImpl<>(list, 0, list.size(), 0, null);
    }

    /**
     * Returns an immutable copy of the elements of the specified {@code collection}. The copy is obtained through the
     * {@code collection}'s {@code toArray} method.
     *
     * @param <E> element type.
     * @param collection collection to be copied.
     * @return an immutable copy of the specified collection.
     * @throws NullPointerException if the specified collection is null.
     */
    static <E> Array<E> copyOf(@Nonnull Collection <E> collection) {
        Objects.requireNonNull(collection);
        @SuppressWarnings("unchecked")
        E[] elements = (E[]) collection.toArray();
        int characteristics = (collection instanceof Set ? (collection instanceof SortedSet ? SORTED | DISTINCT : DISTINCT) : 0);
        return Arrays.newArray(elements, 0, elements.length, characteristics, null, false);
    }

    /**
     * Returns a list backed by the specified array. This method acts as bridge between array-based and collection-based
     * APIs. The returned list is serializable and implements {@link RandomAccess}.
     *
     * <p>If the specified {@code array} is immutable, the returned list is an unmodifiable view of the specified
     * {@code} array. Query operations on the returned list "read through" to the specified {@code array}. Element
     * modification, addition or removal results in an {@code UnsupportedOperationException}.</p>
     *
     * <p>If the specified array is mutable, the returned list is fixed-length. Query operations on the returned list
     * "read through" to the specified {@code array}. Modifying operations on the returned list "write through" to the
     * specified {@code array}. Element addition or removal results in an {@code UnsupportedOperationException}.</p>
     *
     * <p>If the specified array is dynamic, the returned list is a fully functional view of the specified {@code array}.
     * "write through" to the specified {@code array}. Query operations on the returned list "read through" to the
     * specified {@code array}. Modifying, addition and removal operations on the returned list "write through" to the
     * specified {@code array}.
     *
     * @param <E> element type.
     * @param array the array by which the list will be backed.
     * @return a list view of the specified array.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("unchecked")
    static <E> List<E> asList(@Nonnull Array<E> array) {
        Objects.requireNonNull(array);
        if (array instanceof DynamicArray) {
            return new DynamicArrayListImpl<>((DynamicArray<E>) array);
        } else if (array instanceof MutableArray) {
            return new MutableArrayListImpl<>((MutableArray<E>) array);
        } else {
            return new ArrayListImpl<>(array);
        }
    }

    /**
     * Returns a navigable set backed by the specified array. The array can be modified by the navigable set
     * implementation to ensure that it adheres to the semantics of the navigable set. This method acts as bridge
     * between array-based and collection-based APIs. The returned navigable set is serializable.
     *
     * @param <E> element type.
     * @param array the list by which the array will be backed.
     * @return an immutable copy of the specified collection.
     * @throws NullPointerException if the specified collection is null.
     */
    static <E> NavigableSet<E> asNavigableSet(@Nonnull DynamicArray<E> array) {
        Objects.requireNonNull(array);
        return new DynamicArrayNavigableSetImpl<>(array);
    }

    /**
     * Returns a navigable set backed by the specified array. The array can be modified by the navigable set
     * implementation to ensure that it adheres to the semantics of the navigable set. This method acts as bridge
     * between array-based and collection-based APIs. The returned navigable set is serializable.
     *
     * @param <E> element type.
     * @param array the list by which the array will be backed.
     * @param comparator comparator used for sorting set.
     * @return an immutable copy of the specified collection.
     * @throws NullPointerException if the specified collection is null.
     */
    static <E> NavigableSet<E> asNavigableSet(@Nonnull DynamicObjectArray<E> array, @Nullable Comparator<? super E> comparator) {
        Objects.requireNonNull(array);
        return new DynamicArrayNavigableSetImpl<>(array, comparator);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc", "NullableProblems"})
    private static class DynamicArrayListImpl<E> extends AbstractList<E> implements List<E>, RandomAccess, Serializable {
        static final long serialVersionUID = -7791761926378464188L;
        final DynamicArray<E> array;

        public DynamicArrayListImpl(DynamicArray<E> array) {
            this.array = array;
        }

        @SuppressWarnings("unchecked")
        @Override
        public void sort(Comparator<? super E> c) {
            if (c == null) {
                array.sort();
            } else if (array instanceof MutableObjectArray) {
                ((MutableObjectArray) array).sort(c);
            } else {
                Collections.sort(this, c);
            }
        }

        @Override public int size() {
            return array.length();
        }

        @Override public boolean isEmpty() {
            return array.isEmpty();
        }

        @Override public boolean contains(Object o) {
            return array.contains(o);
        }

        @Override public Iterator<E> iterator() {
            return array.iterator();
        }

        @Override public Object[] toArray() {
            return toArray(new Object[size()]);
        }

        @SuppressWarnings("unchecked")
        @Override
        public <T> T[] toArray(T[] a) {
            if (array instanceof ObjectArray) {
                return ((ObjectArray<E>) array).toArray(a);
            } else {
                int length = array.length();
                T[] r = a.length >= length ? a : (T[]) newInstance(a.getClass().getComponentType(), length);
                for (int index = 0; index < length; index++) {
                    r[index] = (T) array.get(index);
                }
                if (a.length > length) {
                    r[length] = null;
                }
                return r;
            }
        }

        @Override public boolean add(E e) {
            array.add(e);
            return true;
        }

        @Override public boolean remove(Object o) {
            return array.remove(o);
        }

        @Override public boolean containsAll(Collection<?> c) {
            return array.containsAll(ArrayBridge.copyOf(c));
        }

        @Override public boolean addAll(Collection<? extends E> c) {
            array.addAll(ArrayBridge.copyOf(c));
            return true;
        }

        @Override
        public boolean addAll(int index, Collection<? extends E> c) {
            array.addAll(index, ArrayBridge.copyOf(c));
            return true;
        }

        @Override
        public boolean removeAll(Collection<?> c) {
            return array.removeAll(ArrayBridge.copyOf(c));
        }

        @Override
        public boolean retainAll(Collection<?> c) {
            return array.retainAll(ArrayBridge.copyOf(c));
        }

        @Override
        public void replaceAll(UnaryOperator<E> operator) {
            array.replaceAll(operator);
        }

        @Override
        public void clear() {
            array.clear();
        }

        @Override
        public E get(int index) {
            return array.get(index);
        }

        @Override
        public E set(int index, E element) {
            return array.set(index, element);
        }

        @Override
        public void add(int index, E element) {
            array.add(index, element);
        }

        @Override
        public E remove(int index) {
            return array.remove(index);
        }

        @Override
        public int indexOf(Object o) {
            return array.indexOf(o);
        }

        @Override
        public int lastIndexOf(Object o) {
            return array.lastIndexOf(o);
        }

        @Override
        public ListIterator<E> listIterator() {
            return array.iterator();
        }

        @Override
        public ListIterator<E> listIterator(int index) {
            return array.iterator(index);
        }

        @Override
        public List<E> subList(int fromIndex, int toIndex) {
            return asList(array.subArray(fromIndex, toIndex));
        }

        @Override
        public Spliterator<E> spliterator() {
            return array.spliterator();
        }

        @Override
        public Stream<E> stream() {
            return array.stream();
        }

        @Override
        public Stream<E> parallelStream() {
            return array.parallelStream();
        }

        @Override
        public void forEach(Consumer<? super E> action) {
            array.forEach(action);
        }
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc", "NullableProblems"})
    private static class MutableArrayListImpl<E> extends AbstractList<E> implements List<E>, RandomAccess, Serializable {
        static final long serialVersionUID = -2570937601815527727L;
        final MutableArray<E> array;

        public MutableArrayListImpl(MutableArray <E> array) {
            this.array = array;
        }

        @SuppressWarnings("unchecked")
        @Override
        public void sort(Comparator<? super E> c) {
            if (array instanceof MutableObjectArray) {
                ((MutableObjectArray) array).sort(c);
            } else {
                Collections.sort(this, c);
            }
        }

        @Override public int size() {
            return array.length();
        }

        @Override public boolean isEmpty() {
            return array.isEmpty();
        }

        @Override public boolean contains(Object o) {
            return array.contains(o);
        }

        @Override public Iterator<E> iterator() {
            return array.iterator();
        }

        @Override public Object[] toArray() {
            return (Object[]) array.toArray();
        }

        @SuppressWarnings("unchecked")
        @Override
        public <T> T[] toArray(T[] a) {
            if (array instanceof ObjectArray) {
                return ((ObjectArray<E>) array).toArray(a);
            } else {
                int length = array.length();
                T[] r = a.length >= length ? a : (T[]) newInstance(a.getClass().getComponentType(), length);
                for (int index = 0; index < length; index++) {
                    r[index] = (T) array.get(index);
                }
                if (a.length > length) {
                    r[length] = null;
                }
                return r;
            }
        }

        @Override public boolean containsAll(Collection<?> c) {
            return array.containsAll(ArrayBridge.copyOf(c));
        }

        @Override
        public void replaceAll(UnaryOperator<E> operator) {
            array.replaceAll(operator);
        }

        @Override
        public E get(int index) {
            return array.get(index);
        }

        @Override
        public E set(int index, E element) {
            return array.set(index, element);
        }

        @Override
        public int indexOf(Object o) {
            return array.indexOf(o);
        }

        @Override
        public int lastIndexOf(Object o) {
            return array.lastIndexOf(o);
        }

        @Override
        public ListIterator<E> listIterator() {
            return array.iterator();
        }

        @Override
        public ListIterator<E> listIterator(int index) {
            return array.iterator(index);
        }

        @Override
        public List<E> subList(int fromIndex, int toIndex) {
            return asList(array.subArray(fromIndex, toIndex));
        }

        @Override
        public Spliterator<E> spliterator() {
            return array.spliterator();
        }

        @Override
        public Stream<E> stream() {
            return array.stream();
        }

        @Override
        public Stream<E> parallelStream() {
            return array.parallelStream();
        }

        @Override
        public void forEach(Consumer<? super E> action) {
            array.forEach(action);
        }
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc", "NullableProblems"})
    private static class ArrayListImpl<E> extends AbstractList<E> implements List<E>, RandomAccess, Serializable {
        static final long serialVersionUID = -8870589484437036889L;
        final Array<E> array;

        public ArrayListImpl(Array<E> array) {
            this.array = array;
        }

        @SuppressWarnings("unchecked")
        @Override
        public void sort(Comparator<? super E> c) {
            if (c == null && array instanceof MutableArray) {
                ((MutableArray) array).sort();
            } else if (array instanceof MutableObjectArray) {
                ((MutableObjectArray) array).sort(c);
            } else {
                Collections.sort(this, c);
            }
        }

        @Override public int size() {
            return array.length();
        }

        @Override public boolean isEmpty() {
            return array.isEmpty();
        }

        @Override public boolean contains(Object o) {
            return array.contains(o);
        }

        @Override public Iterator<E> iterator() {
            return array.iterator();
        }

        @Override public Object[] toArray() {
            return (Object[]) array.toArray();
        }

        @SuppressWarnings("unchecked")
        @Override
        public <T> T[] toArray(T[] a) {
            if (array instanceof ObjectArray) {
                return ((ObjectArray<E>) array).toArray(a);
            } else {
                int length = array.length();
                T[] r = a.length >= length ? a : (T[]) newInstance(a.getClass().getComponentType(), length);
                for (int index = 0; index < length; index++) {
                    r[index] = (T) array.get(index);
                }
                if (a.length > length) {
                    r[length] = null;
                }
                return r;
            }
        }

        @Override
        public E get(int index) {
            return array.get(index);
        }

        @Override
        public int indexOf(Object o) {
            return array.indexOf(o);
        }

        @Override
        public int lastIndexOf(Object o) {
            return array.lastIndexOf(o);
        }

        @Override
        public ListIterator<E> listIterator() {
            return array.iterator();
        }

        @Override
        public ListIterator<E> listIterator(int index) {
            return array.iterator(index);
        }

        @Override
        public List<E> subList(int fromIndex, int toIndex) {
            return asList(array.subArray(fromIndex, toIndex));
        }

        @Override
        public Spliterator<E> spliterator() {
            return array.spliterator();
        }

        @Override
        public Stream<E> stream() {
            return array.stream();
        }

        @Override
        public Stream<E> parallelStream() {
            return array.parallelStream();
        }

        @Override
        public void forEach(Consumer<? super E> action) {
            array.forEach(action);
        }
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static final class ReverseIteratorImpl<E> implements Iterator<E>  {
        final ArrayIterator<E> iterator;

        ReverseIteratorImpl(ArrayIterator<E> iterator) {
            this.iterator = iterator;
        }

        @Override
        public boolean hasNext() {
            return iterator.hasPrevious();
        }

        @Override
        public E next() {
            return iterator.previous();
        }

        @Override
        public void remove() {
            iterator.remove();
        }
    }

    /**
     * Reverses the contents of the specified {@code array}.
     *
     * @param array array which elements need to be put in reversed order.
     */
    public static void reverse(@Nonnull Object[] array) {
        int length = array.length;
        for (int i = 0, mid = length >> 1, j = length - 1; i < mid; i++, j--) {
            Object o = array[i];
            array[i] = array[j];
            array[j] = o;
        }
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc", "NullableProblems"})
    private static class DynamicArrayNavigableSetImpl<E> implements NavigableSet<E>, Serializable {

        // Null restrictions and type compatibility is enforced by by the underlying array and binarySearch function.

        static final long serialVersionUID = 2529740906358890850L;
        final DynamicArray<E> array;
        final Comparator<? super E> comparator;
        final boolean reverse;
        transient ToIntFunction<Object> binarySearchFunction;

        public DynamicArrayNavigableSetImpl(DynamicArray<E> array) {
            this(array, null);
        }

        public DynamicArrayNavigableSetImpl(DynamicArray<E> array, Comparator<? super E> comparator) {
            this(array, comparator, false);
        }

        public DynamicArrayNavigableSetImpl(DynamicArray<E> array, Comparator<? super E> comparator, boolean reverse) {
            assert comparator == null || array instanceof DynamicObjectArray;
            this.array = array;
            this.comparator = comparator;
            this.reverse = reverse;
            this.binarySearchFunction = getBinarySearchFunction(array);
        }

        @SuppressWarnings("unchecked")
        private void distinct() {
            if (comparator == null) {
                array.distinct();
            } else {
                ((DynamicObjectArray) array).distinct(comparator);
            }
        }

        private int getFromIndex(E element, boolean inclusive) {
            int index = binarySearchFunction.applyAsInt(element);
            if (index >= 0) {
                index += (inclusive ? 0 : 1);
            } else {
                index = -index - 1;
            }
            return index;
        }

        private int getToIndex(E element, boolean inclusive) {
            return getFromIndex(element, !inclusive);
        }

        @Override public void clear() {
            array.clear();
        }

        @Override public boolean removeAll(Collection<?> c) {
            distinct();
            return array.removeAll(copyOf(c));
        }

        @Override public boolean retainAll(Collection<?> c) {
            distinct();
            return array.retainAll(copyOf(c));
        }

        @Override public boolean addAll(Collection<? extends E> c) {
            distinct();
            boolean modified = false;
            for (E e : c) {
                int index = binarySearchFunction.applyAsInt(e);
                if (index < 0) {
                    array.add(-index - 1, e);
                    modified = true;
                }
            }
            if (!array.hasCharacteristics(SORTED | DISTINCT)) { // Array implementation might have removed characteristics as result of add call.
                array.doAction(context -> context.addCharacteristics(SORTED | DISTINCT));
            }
            return modified;
        }

        @Override public boolean containsAll(Collection <?> c) {
            distinct();
            return array.containsAll(copyOf(c));
        }

        @Override public boolean remove(Object o) {
            distinct();
            return array.remove(o);
        }

        @SuppressWarnings("unchecked")
        @Override public boolean add(E e) {
            distinct();
            int index = binarySearchFunction.applyAsInt(e);
            boolean modified = false;
            if (index < 0) {
                array.add(-index - 1, e);
                modified = true;
            }
            if (!array.hasCharacteristics(SORTED | DISTINCT)) { // Array implementation might have removed characteristics as result of add call.
                array.doAction(context -> context.addCharacteristics(SORTED | DISTINCT));
            }
            return modified;
        }

        @SuppressWarnings("unchecked")
        @Override public <T> T[] toArray(T[] a) {
            distinct();
            T[] r;
            if (array instanceof ObjectArray) {
                r = ((ObjectArray<E>) array).toArray(a);
                if (reverse) {
                    reverse(r);
                }
            } else {
                int length = array.length();
                r = a.length >= length ? a : (T[]) newInstance(a.getClass().getComponentType(), length);
                Object o = array.toArray();
                for (int index = 0; index < length; index++) {
                    r[index] = (T) get(o, reverse ? length - index - 1 : index);
                }
                if (a.length > length) {
                    r[length] = null;
                }
            }
            return r;
        }

        @Override public Object[] toArray() {
            return toArray(new Object[size()]);
        }

        @Override public boolean contains(Object o) {
            distinct();
            return array.contains(o);
        }

        @Override public boolean isEmpty() {
            return array.isEmpty();
        }

        @Override public Iterator<E> iterator() {
            distinct();
            return reverse ? new ReverseIteratorImpl<>(array.iterator(array.length())) : array.iterator();
        }

        @Override public int size() {
            distinct();
            return array.length();
        }

        @Override public E lower(E e) {
            return reverse ? higher0(e) : lower0(e);
        }

        private E lower0(E e) {
            distinct();
            int index = binarySearchFunction.applyAsInt(e);
            if (index < -1) {
                return array.get(-index - 2);
            }
            if (index > 0) {
                return array.get(index - 1);
            }
            return null;
        }

        @Override
        public E floor(E e) {
            return reverse ? ceiling0(e) : floor0(e);
        }

        private E floor0(E e) {
            distinct();
            int index = binarySearchFunction.applyAsInt(e);
            if (index < -1) {
                return array.get(-index - 2);
            }
            if (index >= 0) {
                return array.get(index);
            }
            return null;
        }

        @Override
        public E ceiling(E e) {
            return reverse ? floor0(e) : ceiling0(e);
        }

        private E ceiling0(E e) {
            distinct();
            int index = binarySearchFunction.applyAsInt(e);
            if (index < 0 && -index - 1 < array.length()) {
                return array.get(-index - 1);
            }
            if (index >= 0) {
                return array.get(index);
            }
            return null;
        }

        @Override
        public E higher(E e) {
            return reverse ? lower0(e) : higher0(e);
        }

        private E higher0(E e) {
            distinct();
            int index = binarySearchFunction.applyAsInt(e);
            if (index < 0 && -index -1 < array.length()) {
                return array.get(-index - 1);
            }
            if (index >= 0 && index + 1 < array.length()) {
                return array.get(index + 1);
            }
            return null;
        }

        @Override public E pollFirst() {
            return reverse ? pollLast0() : pollFirst0();
        }

        private E pollFirst0() {
            distinct();
            if (array.length() > 0) {
                return array.remove(0);
            }
            return null;
        }

        @Override
        public E pollLast() {
            return reverse ? pollFirst0() : pollLast0();
        }

        private E pollLast0() {
            distinct();
            int length = array.length();
            if (length > 0) {
                return array.remove(length - 1);
            }
            return null;
        }

        @Override public NavigableSet<E> descendingSet() {
            return new DynamicArrayNavigableSetImpl<>(array, comparator, true);
        }

        @Override public Iterator<E> descendingIterator() {
            distinct();
            return reverse ? array.iterator() : new ReverseIteratorImpl<>(array.iterator(array.length()));
        }

        @Override public NavigableSet<E> subSet(E fromElement, boolean fromInclusive, E toElement, boolean toInclusive) {
            return reverse ? subSet0(toElement, toInclusive, fromElement, fromInclusive) : subSet0(fromElement, fromInclusive, toElement, toInclusive);
        }

        public NavigableSet<E> subSet0(E fromElement, boolean fromInclusive, E toElement, boolean toInclusive) {
            distinct();
            int fromIndex = getFromIndex(fromElement, fromInclusive);
            int toIndex = getToIndex(toElement, toInclusive);
            if (fromIndex > toIndex) {
                throw new IllegalArgumentException(String.format("fromIndex (%d) > toIndex (%d)", fromIndex, toIndex));
            }
            return new DynamicArrayNavigableSetImpl<>(array.subArray(fromIndex, toIndex), comparator, reverse);
        }

        @Override
        public NavigableSet<E> headSet(E toElement, boolean inclusive) {
            return reverse ? tailSet0(toElement, inclusive) : headSet0(toElement, inclusive);
        }

        private NavigableSet<E> headSet0(E toElement, boolean inclusive) {
            distinct();
            return new DynamicArrayNavigableSetImpl<>(array.length(getToIndex(toElement, inclusive)), comparator, reverse);
        }

        @Override
        public NavigableSet<E> tailSet(E fromElement, boolean inclusive) {
            return reverse ? headSet0(fromElement, inclusive) : tailSet0(fromElement, inclusive);
        }

        private NavigableSet<E> tailSet0(E fromElement, boolean inclusive) {
            distinct();
            return new DynamicArrayNavigableSetImpl<>(array.offset(getFromIndex(fromElement, inclusive)), comparator, reverse);
        }

        @Override public SortedSet<E> subSet(E fromElement, E toElement) {
            return subSet(fromElement, true, toElement, false);
        }

        @Override
        public SortedSet<E> headSet(E toElement) {
            return headSet(toElement, false);
        }

        @Override
        public SortedSet<E> tailSet(E fromElement) {
            return tailSet(fromElement, true);
        }

        @Override public Comparator<? super E> comparator() {
            return array.getComparator();
        }

        @Override
        public E first() {
            return reverse ? last0() : first0();
        }

        private E first0() {
            distinct();
            return array.length() > 0 ? array.get(0) : null;
        }

        @Override
        public E last() {
            return reverse ? first0() : last0();
        }

        private E last0() {
            distinct();
            int length;
            if ((length = array.length()) > 0) {
                return array.get(length - 1);
            }
            return null;
        }

        @Override
        public int hashCode() {
            return array.hashCode();
        }

        @Override
        public boolean equals(Object obj) {
            if (obj == this) {
                return true;
            }
            if (!(obj instanceof Set)) {
                return false;
            }
            Collection<?> c = (Collection<?>) obj;
            if (c.size() != size()) {
                return false;
            }
            try {
                return containsAll(c);
            } catch (ClassCastException unused)   {
                return false;
            } catch (NullPointerException unused) {
                return false;
            }
        }

        @Override
        public String toString() {
            return array.toString();
        }

        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            binarySearchFunction = getBinarySearchFunction(array);
        }
    }

    /**
     * Creates a function that looks up the {@code Arrays.unsafeBinarySearch} method that matches the type of the
     * specified array. This function can be re-used by the caller to prevent unnecessary lookups.
     *
     * <p>The semantics of the function are as follows:</p>
     *
     * <p>Searches the specified array for the specified object (function argument) using the binary search algorithm.
     * The array must be sorted into ascending order according to the specified comparator (as by the
     * {@link java.util.Arrays#sort(Object[], Comparator) sort(T[], Comparator)} method) prior to making this call. If
     * it is not sorted, the results are undefined. If the array contains multiple elements equal to the specified
     * object, there is no guarantee which one will be found. Unlike {@code Arrays.binarySearch}, no exception is thrown
     * if specified array is not marked as sorted.</p>
     *
     * @param array the array to be searched.
     * @return function that implements binary search algorithm.
     */
    private static ToIntFunction<Object> getBinarySearchFunction(@Nonnull Array<?> array) {
        if (array instanceof BooleanArray) {
            return key -> key instanceof Boolean ? Arrays.binarySearch((BooleanArray) array, (Boolean) key) : array.indexOf(key);
        } else if (array instanceof ByteArray) {
            return key -> key instanceof Byte ? Arrays.binarySearch((ByteArray) array, (Byte) key) : array.indexOf(key);
        } else if (array instanceof CharArray) {
            return key -> key instanceof Character ? Arrays.binarySearch((CharArray) array, (Character) key) : array.indexOf(key);
        } else if (array instanceof ShortArray) {
            return key -> key instanceof Short ? Arrays.binarySearch((ShortArray) array, (Short) key) : array.indexOf(key);
        } else if (array instanceof IntArray) {
            return key -> key instanceof Integer ? Arrays.binarySearch((IntArray) array, (Integer) key) : array.indexOf(key);
        } else if (array instanceof LongArray) {
            return key -> key instanceof Long ? Arrays.binarySearch((LongArray) array, (Long) key) : array.indexOf(key);
        } else if (array instanceof FloatArray) {
            return key -> key instanceof Float ? Arrays.binarySearch((FloatArray) array, (Float) key) : array.indexOf(key);
        } else if (array instanceof DoubleArray) {
            return key -> key instanceof Double ? Arrays.binarySearch((DoubleArray) array, (Double) key) : array.indexOf(key);
        } else if (array instanceof ObjectArray) {
            @SuppressWarnings("unchecked")
            ObjectArray<Object> a = (ObjectArray<Object>) array;
            return key -> Arrays.binarySearch(a, key, a.getComparator());
        } else {
            return array::indexOf;
        }
    }
}
