/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License a with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import org.squarebrackets.function.ByteConsumer;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import javax.annotation.concurrent.Immutable;
import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Comparator;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.function.Consumer;

import static java.lang.System.arraycopy;
import static java.util.Arrays.copyOfRange;

/**
 * <p>Immutable array of byte primitives.</p>
 *
 * <p>This byte array is constructed by the static interface methods of {@code ByteArray}. The {@code copyOf}
 * method returns a new array representing a shallow copy of the (native or framework) array argument. The 
 * {@code unsafeValueOf} methods wrap a new {@code ByteArray} around a native array. It is imperative that this 
 * native array is not modified outside the scope of this class, as this would break the immutable contract. 
 * The {@code unsafeValueOf} method is used to prevent unnecessary array copies. It is the caller's responsibility not 
 * to violate the contract.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.</p>
 *
 * <p>This class attempts to establish these characteristics to accommodate these optimizations at best effort. This 
 * means that this implementation determines which set of characteristics apply to its elements without traversing
 * these elements, in other words, without incurring a performance penalty. Primitive arrays, such as this class, 
 * always report {@code NONNULL}. The set of characteristics of immutable arrays are constant.</p>
 * 
 * @author Leon van Zantvoort
 */
@Immutable
class ByteArrayImpl implements ByteArray, NativeArrayProvider, Serializable {

    /** Backing array. */
    transient byte[] elements;

    /** Start index of range. */
    final int offset;

    /** Length of range. */
    int length;

    /** Array characteristics. */
    int characteristics;

    /**
     * Package visible constructor, to be used by static factory methods only.
     *
     * @param array backing array.
     * @param offset start index of range.
     * @param length length of range.
     * @param characteristics array characteristics.
     * @param copy {@code true} to make a defensive copy of {@code array}, {@code false} to use {@code array} reference
     *             as backing array.
     */
    ByteArrayImpl(@Nonnull byte[] array, int offset, int length, int characteristics, boolean copy) {
        Objects.requireNonNull(array);
        if (offset < 0 || length < 0 || length > array.length - offset) {
            throw new IndexOutOfBoundsException();
        }
        if (copy) {
            this.elements = new byte[length];
            arraycopy(array, offset, elements, 0, length);
            this.offset = 0;
        } else {
            this.elements = array;
            this.offset = offset;
        }
        this.length = length;
        this.characteristics = filterCharacteristics(characteristics) | (length > 1 ? 0 : SORTED | DISTINCT);
    }

    /**
     * Checks if the given index is in range. If not, throws an appropriate runtime exception.
     *
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    final void rangeCheck(int index) {
        Arrays.rangeCheck(length, index);
    }

    /**
     * Checks if the given boundary is within this array's boundaries.
     *
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    final void subArrayCheck(int offset, int length) {
        Arrays.subArrayCheck(this.offset, this.length, offset, length);
    }

    /**
     * Returns the component type of the backing array.
     *
     * @return component type of the backing array.
     */
    @Override
    public final Class<Byte> getComponentType() {
        return Byte.TYPE;
    }

    /**
     * Gets the comparator that is used for sorting this array, {@code null} if array is primitive array, or if array
     * sorted according to natural ordering of elements, or {@code null} if array is not sorted.
     *
     * @return a {@code Comparator}, or {@code null} if the elements are sorted in the natural order, or {@code null} if
     * array is not sorted.
     */
    @Override
    public final @CheckForNull Comparator<? super Byte> getComparator() {
        return null;
    }

    /**
     * Returns {@code true} if the native array returned by {@code nativeArray()} is backing the implementing array,
     * {@code false} if the native array is a shallow copy of elements.
     *
     * @return {@code true} if the native array is backing the implementing array, {@code false} otherwise.
     */
    @Override
    public final boolean isBackingArray() {
        return true;
    }

    /**
     * <p>Returns a native array containing all elements of this array. The elements are stored starting from the
     * offset of the implementing array. The total length of this arrays is {@literal >= offset + length} of the
     * implementing array.
     *
     * <p>The resulting object must be cast to the primitive or object type of the implementing array's component type.</p>
     *
     * @return native array.
     */
    @Override
    public final byte[] nativeArray() {
        return elements;
    }

    /**
     * Allow array implementation to filter characteristics based on array specific logic.
     *
     * <p>Default implementation passes specified value straight through.</p>
     *
     * @param characteristics characteristics to be filtered.
     * @return the filtered characteristics.
     */
    protected int filterCharacteristics(int characteristics) {
        return characteristics | NONNULL & ~(MUTABLE | DYNAMIC);
    }

    /**
     * Returns the characteristics set for this array. The result is represented as ORed values from
     * {@link #NONNULL}, {@link #MUTABLE}, {@link #DYNAMIC}, {@link #SORTED}, {@link #DISTINCT}.
     * Modification of the array contents may result in updated characteristics.
     *
     * @return the characteristics set for this array.
     */
    @Override
    public final int characteristics() {
        return characteristics;
    }

    /**
     * <p>Returns the offset relative to the backing array.</p>
     *
     * @return offset relative to the backing array.
     */
    @Override
    public final int offset() {
        return offset;
    }

    /**
     * <p>Returns the number of elements of this array. All index related methods must respect the boundary represented
     * by this value.</p>
     *
     * <p>Note that for mutable arrays, the backing array - as returned by {@code MutableArray.array()} -
     * may be larger.</p>
     *
     * @return length of array.
     */
    @Override
    public final int length() {
        return length;
    }

    /**
     * Returns {@code true} if this array contains no elements.
     *
     * @return {@code true} if this array contains no elements.
     */
    @Override
    public final boolean isEmpty() {
        return length == 0;
    }

    /**
     * Returns the object found at {@code [offset + index]} of the backing array.
     *
     * @param index index relative to the offset of the array.
     * @return object found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @Override
    public Byte get(int index) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        return getByte(index);
    }

    /**
     * Returns the byte value found at {@code [offset + index]} of the backing array.
     *
     * @param index index relative to the offset of the array.
     * @return byte value found at the specified index.
     * @throws IndexOutOfBoundsException if {@code index} position is invalid.
     */
    @Override
    public byte getByte(int index) {
        rangeCheck(index);
        return elements[offset + index];
    }

    /**
     * Returns {@code true} if this array contains the specified object. This method supports boxed primitives. This
     * means that if this array represents a primitive array, this method returns {@code true} if this array contains
     * the unboxed version of specified object.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object object whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified object.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public boolean contains(Object object) {
        if (object instanceof Byte) {
            if (Tripwire.ENABLED) {
                Tripwire.trip(getClass());
            }
            return containsByte((Byte) object);
        }
        return false;
    }

    /**
     * Returns {@code true} if this array contains the specified value.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param value value whose presence in this array is to be tested.
     * @return {@code true} if this array contains the specified value.
     */
    @Override
    public boolean containsByte(byte value) {
        if (hasCharacteristics(SORTED)) {
            return Arrays.binarySearch(this, value) >= 0;
        }
        final int toIndex = offset + length;
        for (int index = offset; index < toIndex; index++) {
            if (elements[index] == value) {
                return true;
            }
        }
        return false;
    }

    /**
     * Returns {@code true} if this array contains all of the elements in the specified {@code array}. Primitives and
     * their object counterparts are regarded equal in this respect.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array is sorted, otherwise all elements are
     * traversed until the objects in specified array are found.</p>
     *
     * @param array array to be checked for containment in this array.
     * @return {@code true} if this array contains all of the elements in the specified array.
     * @throws NullPointerException if array is null.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public boolean containsAll(@Nonnull Array<?> array) {
        ByteArray cast = ArrayCast.toByteArray(array);
        if (cast != null) {
            final byte[] keys = NativeArrayProvider.getNativeByteArray(cast);
            final int fromIndex = array.offset();
            final int toIndex = fromIndex + array.length();
            for (int index = fromIndex; index < toIndex; index++) {
                if (!containsByte(keys[index])) {
                    return false;
                }
            }
            return true;
        }
        for (Object o : array) {
            if (!(o instanceof Byte) || !containsByte((Byte) o)) {
                return false;   // Returns false for null elements.
            }
        }
        return true;
    }

    /**
     * Returns the index of the first occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified object is found.</p>
     *
     * @param object whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public int indexOf(Object object) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        if (object instanceof Byte) {
            return indexOfByte((Byte) object);
        }
        return -1;
    }

    /**
     * Returns the index of the first occurrence of the specified value in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed until the specified value is found.</p>
     *
     * @param value value whose lowest index in this array is to be returned.
     * @return lowest index of value in array or {@code -1} if value is not found.
     */
    @Override
    public int indexOfByte(byte value) {
        if (hasCharacteristics(SORTED)) {
            int index = Arrays.binarySearch(this, value);
            if (index >= 0) {
                return Arrays.binarySearchFirst(this, index);
            }
            return -1;
        }
        final byte[] a = elements;
        final int toIndex = offset + length;
        for (int index = offset; index < toIndex; index++) {
            if (a[index] == value) {
                return index - offset;
            }
        }
        return -1;
    }

    /**
     * Returns the index of the last occurrence of the specified object in the array, or {@code -1} if
     * the array does not contain the object. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified object is found.</p>
     *
     * @param object object whose highest index in this array is to be returned.
     * @return highest index of object in array or {@code -1} if value is not found.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    public int lastIndexOf(Object object) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        if (object instanceof Byte) {
            return lastIndexOfByte((Byte) object);
        }
        return -1;
    }

    /**
     * Returns the index of the last occurrence of the specified value in the array, or {@code -1} if
     * the array does not contain the value. The index is relative to the offset of the array.
     *
     * <p>The Java Array Framework implementation performs a binary search if this array reports {@code SORTED}
     * characteristics, otherwise all elements are traversed backwards until the specified value is found.</p>
     *
     * @param value value whose highest index in this array is to be returned.
     * @return highest index of value in array or {@code -1} if value is not found.
     */
    @Override
    public int lastIndexOfByte(byte value) {
        if (hasCharacteristics(SORTED)) {
            int index = Arrays.binarySearch(this, value);
            if (index >= 0) {
                return Arrays.binarySearchLast(this, index);
            }
            return -1;
        }
        final byte[] a = elements;
        final int toIndex = offset + length;
        for (int index = toIndex -1; index >= offset; index--) {
            if (a[index] == value) {
                return index - offset;
            }
        }
        return -1;
    }

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within this array.
     *
     * @param consumer consumer that is called for each value within this array.
     * @throws NullPointerException if consumer is null.
     */
    @Override
    public void forEach(@Nonnull Consumer<? super Byte> consumer) {
        if (Tripwire.ENABLED) {
            Tripwire.trip(getClass());
        }
        forEachByte(consumer::accept);
    }

    /**
     * Iterates through the array and invokes the consumer's {@code accept} method for each value within this array.
     *
     * @param consumer consumer that is called for each value within this array.
     * @throws NullPointerException if consumer is null.
     */
    @Override
    public void forEachByte(@Nonnull ByteConsumer consumer) {
        Objects.requireNonNull(consumer);
        final byte[] a = elements;
        final int toIndex = offset + length;
        for (int index = offset; index < toIndex; index++) {
            consumer.accept(a[index]);
        }
    }

    /**
     * Returns an native array of the elements specified by {@code this} array. This method always returns a defensive
     * (range) copy of the backing array, even if this array is mutable.
     *
     * @return range copy of backing array.
     */
    @Override
    public byte[] toArray() {
        return copyOfRange(elements, offset, offset + length);
    }

    /**
     * Returns an array containing all of the byte values of {@code this} array. If the byte values fit in the
     * specified array, it is returned therein. Otherwise, a new byte array is allocated.
     *
     * @param array the array into which the elements of this array are to be stored, if it is big enough; otherwise, a
     *              new array of the same runtime type is allocated for this purpose.
     * @return an array containing all of the elements in the range specified by this array.
     * @throws NullPointerException if the specified array is null.
     */
    @Override
    public byte[] toArray(@Nonnull byte[] array) {
        byte[] a = array.length >= length ? array : new byte[length];
        arraycopy(elements, offset, a, 0, length);
        return a;
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    public ByteArray offset(int offset) {
        if (offset == 0) {
            return this;
        }
        int newOffset = this.offset + offset;
        int newLength = length - offset;
        subArrayCheck(newOffset, newLength);
        return new ByteArrayImpl(elements, newOffset, newLength, characteristics, false);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    public ByteArray length(int length) {
        if (this.length == length) {
            return this;
        }
        subArrayCheck(offset, length);
        return new ByteArrayImpl(elements, offset, length, characteristics, false);
    }

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    public ByteArray subArray(int fromIndex, int toIndex) {
        int newOffset = offset + fromIndex;
        int newLength = toIndex - fromIndex;
        if (newOffset == offset && newLength == length) {
            return this;
        }
        subArrayCheck(newOffset, newLength);
        return new ByteArrayImpl(elements, newOffset, newLength, characteristics, false);
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private class ArrayIteratorImpl implements ArrayIterator.OfByte {
        int index = offset;

        ArrayIteratorImpl() {
            this(0);
        }

        ArrayIteratorImpl(int index) {
            this.index = offset + index;
        }

        @Override
        public boolean hasNext() {
            return index < offset + length;
        }

        @Override
        public byte nextByte() {
            int i = index;
            if (i >= offset + length) {
                throw new NoSuchElementException();
            }
            assert i < elements.length;
            index = i + 1;
            return elements[i];
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }

        @Override
        public boolean hasPrevious() {
            return index > offset;
        }

        @Override
        public byte previousByte() {
            int i = index - 1;
            if (i < offset) {
                throw new NoSuchElementException();
            }
            assert i < elements.length;
            index = i;
            return elements[i];
        }

        @Override
        public int nextIndex() {
            return index;
        }

        @Override
        public int previousIndex() {
            return index - 1;
        }

        @Override
        public void setByte(byte value) {
            throw new UnsupportedOperationException();
        }

        @Override
        public void addByte(byte value) {
            throw new UnsupportedOperationException();
        }
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @return an array iterator over the elements in this array.
     */
    @Override
    public ArrayIterator.OfByte iterator() {
        return new ArrayIteratorImpl();
    }

    /**
     * Returns an array iterator over the elements in this array.
     *
     * @param index index of the first element to be returned from the array iterator.
     * @return a array iterator over the elements in this array.
     * @throws IndexOutOfBoundsException if the index is out of range.
     */
    @Override
    public ArrayIterator.OfByte iterator(int index) {
        if (index > length) {
            throw new IndexOutOfBoundsException();
        }
        return new ArrayIteratorImpl(index);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public String toString() {
        int iMax = offset + length - 1;
        if (iMax - offset == -1) {
            return "[]";
        }
        StringBuilder b = new StringBuilder();
        b.append('[');
        for (int i = offset; ; i++) {
            b.append(elements[i]);
            if (i == iMax) {
                return b.append(']').toString();
            }
            b.append(", ");
        }
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public int hashCode() {
        int result = 1;
        int to = offset + length;
        for (int i = offset; i < to; i++) {
            byte element = elements[i];
            //noinspection ShiftOutOfRange
            int elementHash = (element ^ (element >>> 32));
            result = 31 * result + elementHash;
        }
        return result;
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }
        if (obj instanceof Array) {
            ByteArray a = ArrayCast.toByteArray((Array) obj);
            if (a != null) {
                byte[] a1 = elements;
                byte[] a2 = NativeArrayProvider.getNativeByteArray(a);
                int l2 = a.length();
                if (length != l2) {
                    return false;
                }
                int o1 = offset;
                int o2 = a.offset();
                for (int i = 0; i < length; i++) {
                    if (a1[o1 + i] != a2[o2 + i]) {
                        return false;
                    }
                }
                return true;
            }
            return Arrays.equals(this, (Array) obj);
        }
        return false;
    }

    /**
     * Write replace method for the serialization proxy pattern. This method must be overwritten by sub classes if their
     * serialized form differs from this class.
     *
     * @return serialization proxy.
     */
    Object writeReplace() {
        return new SerializationProxy(this);
    }

    @SuppressWarnings("MissingMethodJavaDoc")
    private void readObject(ObjectInputStream s) throws InvalidObjectException {
        throw new InvalidObjectException("Proxy required");
    }

    @SuppressWarnings({"MissingMethodJavaDoc", "MissingFieldJavaDoc"})
    private static class SerializationProxy implements Serializable {
        static final long serialVersionUID = 7505111204405719668L;

        transient int offset;
        final int length;
        final int characteristics;
        transient byte[] elements;

        SerializationProxy(ByteArrayImpl array) {
            this.offset = array.offset;
            this.length = array.length;
            this.characteristics = array.characteristics;
            this.elements = array.elements;
        }

        private void writeObject(ObjectOutputStream s) throws IOException {
            s.defaultWriteObject();
            for (int i = 0; i < length; i++) {
                s.writeByte(elements[offset + i]);
            }
        }

        private void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
            s.defaultReadObject();
            elements = new byte[length];
            for (int i = 0; i < length; i++) {
                elements[i] = s.readByte();
            }
        }

        private Object readResolve() {
            return new ByteArrayImpl(elements, 0, length, characteristics, false);
        }
    }
}

