/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import java.util.Iterator;

/**
 * Arrays implementing this interface provide fast access to a native array containing all elements, including elements
 * of parent arrays. This interface allows utility classes such as {@link Arrays} to perform fast read operations on
 * native arrays. The native arrays returned by this interface must only use for read access. Writing is prohibited.
 *
 * @author Leon van Zantvoort
 */
interface NativeArrayProvider {

    /**
     * Returns {@code true} if the array returned by {@code nativeArray()} is the backing array of the specified
     * {@code array}, {@code false} otherwise.
     *
     * @param array array to be inspected.
     * @param componentType component type to test for.
     * @return {@code true} if the array returned by {@code nativeArray()} is the backing array of the specified
     * {@code array}, {@code false} otherwise.
     */
    static boolean isBackingArray(Array<?> array, Class<?> componentType) {
        return array instanceof NativeArrayProvider && ((NativeArrayProvider) array).isBackingArray() &&
                componentType.isAssignableFrom(array.getComponentType());
    }

    /**
     * Returns native array of the specified {@code array}.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull boolean[] getNativeBooleanArray(Array<? extends Boolean> array) {
        if (array.getComponentType() == Boolean.TYPE && array instanceof NativeArrayProvider) {
            return (boolean[]) ((NativeArrayProvider) array).nativeArray();
        }
        if (array instanceof BooleanArray) {
            final int offset = array.offset();
            final int length = array.length();
            boolean[] a = ((BooleanArray) array).toArray(new boolean[offset + length]);
            System.arraycopy(a, 0, a, offset, length);
            return a;
        }
        final int fromIndex = array.offset();
        final int toIndex = fromIndex + array.length();
        final boolean[] a = new boolean[toIndex];
        final Iterator<? extends Boolean> iterator = array.iterator(fromIndex);
        for (int index = fromIndex; index < toIndex; index++) {
            a[index] = iterator.next(); // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns native array of the specified {@code array}.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull byte[] getNativeByteArray(Array<? extends Number> array) {
        if (array.getComponentType() == Byte.TYPE && array instanceof NativeArrayProvider) {
            return (byte[]) ((NativeArrayProvider) array).nativeArray();
        }
        if (array instanceof ByteArray) {
            final int offset = array.offset();
            final int length = array.length();
            byte[] a = ((ByteArray) array).toArray(new byte[offset + length]);
            System.arraycopy(a, 0, a, offset, length);
            return a;
        }
        final int fromIndex = array.offset();
        final int toIndex = fromIndex + array.length();
        final byte[] a = new byte[toIndex];
        final Iterator<? extends Number> iterator = array.iterator(fromIndex);
        for (int index = fromIndex; index < toIndex; index++) {
            a[index] = iterator.next().byteValue(); // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns native array of the specified {@code array}.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull char[] getNativeCharArray(Array<? extends Character> array) {
        if (array.getComponentType() == Character.TYPE && array instanceof NativeArrayProvider) {
            return (char[]) ((NativeArrayProvider) array).nativeArray();
        }
        if (array instanceof CharArray) {
            final int offset = array.offset();
            final int length = array.length();
            char[] a = ((CharArray) array).toArray(new char[offset + length]);
            System.arraycopy(a, 0, a, offset, length);
            return a;
        }
        final int fromIndex = array.offset();
        final int toIndex = fromIndex + array.length();
        final char[] a = new char[toIndex];
        final Iterator<? extends Character> iterator = array.iterator(fromIndex);
        for (int index = fromIndex; index < toIndex; index++) {
            a[index] = iterator.next(); // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns native array of the specified {@code array}.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull short[] getNativeShortArray(Array<? extends Number> array) {
        if (array.getComponentType() == Short.TYPE && array instanceof NativeArrayProvider) {
            return (short[]) ((NativeArrayProvider) array).nativeArray();
        }
        if (array instanceof ShortArray) {
            final int offset = array.offset();
            final int length = array.length();
            short[] a = ((ShortArray) array).toArray(new short[offset + length]);
            System.arraycopy(a, 0, a, offset, length);
            return a;
        }
        final int fromIndex = array.offset();
        final int toIndex = fromIndex + array.length();
        short[] a = new short[toIndex];
        final Iterator<? extends Number> iterator = array.iterator(fromIndex);
        for (int index = fromIndex; index < toIndex; index++) {
            a[index] = iterator.next().shortValue();    // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns native array of the specified {@code array}.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull int[] getNativeIntArray(Array<? extends Number> array) {
        if (array.getComponentType() == Integer.TYPE && array instanceof NativeArrayProvider) {
            return (int[]) ((NativeArrayProvider) array).nativeArray();
        }
        if (array instanceof IntArray) {
            final int offset = array.offset();
            final int length = array.length();
            int[] a = ((IntArray) array).toArray(new int[offset + length]);
            System.arraycopy(a, 0, a, offset, length);
            return a;
        }
        final int fromIndex = array.offset();
        final int toIndex = fromIndex + array.length();
        final int[] a = new int[toIndex];
        final Iterator<? extends Number> iterator = array.iterator(fromIndex);
        for (int index = fromIndex; index < toIndex; index++) {
            a[index] = iterator.next().intValue();  // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns native array of the specified {@code array}.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull long[] getNativeLongArray(Array<? extends Number> array) {
        if (array.getComponentType() == Long.TYPE && array instanceof NativeArrayProvider) {
            return (long[]) ((NativeArrayProvider) array).nativeArray();
        }
        if (array instanceof LongArray) {
            final int offset = array.offset();
            final int length = array.length();
            long[] a = ((LongArray) array).toArray(new long[offset + length]);
            System.arraycopy(a, 0, a, offset, length);
            return a;
        }
        final int fromIndex = array.offset();
        final int toIndex = fromIndex + array.length();
        final long[] a = new long[toIndex];
        final Iterator<? extends Number> iterator = array.iterator(fromIndex);
        for (int index = fromIndex; index < toIndex; index++) {
            a[index] = iterator.next().longValue(); // Throws NullPointerException if element is null.
        }
        return a;
    }

    /**
     * Returns native array of the specified {@code array}.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull float[] getNativeFloatArray(Array<? extends Number> array) {
        if (array.getComponentType() == Float.TYPE && array instanceof NativeArrayProvider) {
            return (float[]) ((NativeArrayProvider) array).nativeArray();
        }
        if (array instanceof FloatArray) {
            final int offset = array.offset();
            final int length = array.length();
            float[] a = ((FloatArray) array).toArray(new float[offset + length]);
            System.arraycopy(a, 0, a, offset, length);
            return a;
        }
        final int fromIndex = array.offset();
        final int toIndex = fromIndex + array.length();
        final float[] a = new float[toIndex];
        final Iterator<? extends Number> iterator = array.iterator(fromIndex);
        for (int index = fromIndex; index < toIndex; index++) {
            a[index] = iterator.next().floatValue(); // Throws NullPointerException if element is null.
        }
        return a;
    }

    /**
     * Returns native array of the specified {@code array}.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull double[] getNativeDoubleArray(Array<? extends Number> array) {
        if (array.getComponentType() == Double.TYPE && array instanceof NativeArrayProvider) {
            return (double[]) ((NativeArrayProvider) array).nativeArray();
        }
        if (array instanceof DoubleArray) {
            final int offset = array.offset();
            final int length = array.length();
            double[] a = ((DoubleArray) array).toArray(new double[offset + length]);
            System.arraycopy(a, 0, a, offset, length);
            return a;
        }
        final int fromIndex = array.offset();
        final int toIndex = fromIndex + array.length();
        final double[] a = new double[toIndex];
        final Iterator<? extends Number> iterator = array.iterator(fromIndex);
        for (int index = fromIndex; index < toIndex; index++) {
            a[index] = iterator.next().doubleValue();   // Throws NullPointerException for null elements.
        }
        return a;
    }

    /**
     * Returns native array of the specified {@code array}.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    @SuppressWarnings("unchecked")
    static @Nonnull Object[] getNativeArray(Array<?> array) {
        if (!array.getComponentType().isPrimitive() && array instanceof NativeArrayProvider) {
            return (Object[]) ((NativeArrayProvider) array).nativeArray();
        }
        if (array instanceof ObjectArray) {
            final int offset = array.offset();
            final int length = array.length();
            Object[] a = ((ObjectArray) array).toArray(new Object[offset + length]);
            System.arraycopy(a, 0, a, offset, length);
            return a;
        }
        final int fromIndex = array.offset();
        final int toIndex = fromIndex + array.length();
        final Object[] a = new Object[toIndex];
        final Iterator<?> iterator = array.iterator(fromIndex);
        for (int index = fromIndex; index < toIndex; index++) {
            a[index] = iterator.next();
        }
        return a;
    }

    /**
     * Returns {@code true} if the native array returned by {@code nativeArray()} is backing the implementing array,
     * {@code false} if the native array is a shallow copy of elements.
     *
     * @return {@code true} if the native array is backing the implementing array, {@code false} otherwise.
     */
    boolean isBackingArray();

    /**
     * <p>Returns a native array containing all elements of this array. The elements are stored starting from the
     * offset of the implementing array. The total length of this arrays is {@literal >= offset + length} of the
     * implementing array.
     *
     * <p>The resulting object must be cast to the primitive or object type of the implementing array's component type.</p>
     *
     * @return native array.
     */
    Object nativeArray();
}
